Documentation
¶
Index ¶
Constants ¶
const AttributeNameTextLine = api.AttributeName("text-line")
AttributeNameTextLine is the attribute name for individual text lines accessed via the Line embedded accessor.
const NamespaceProperty = "configHub.kubernetes.namespace"
This is also defined in the bridge.
Variables ¶
This section is empty.
Functions ¶
func RegisterFunctions ¶
func RegisterFunctions(rp *textkit.TextResourceProviderType, fh handler.FunctionRegistry)
RegisterFunctions registers all Text functions onto the provided FunctionHandler using the given registrar's resource provider.
Types ¶
This section is empty.
Click to show internal directories.
Click to hide internal directories.